// Copyright (c) Renewed Vision, LLC. All rights reserved.

#ifndef PROCORE_INCLUDE_ACTION_HANDLER_CONTEXTS_INL
#error Don't include TriggerEngine.inl directly; include TriggerEngine.h instead
#endif

#include "MessageLog.h"
#include "TimerHelpers.h"
#include "TriggerEngine.h"

template<typename F>
void pro::ActionResolveContext::RequestMacro(const pro_data::ResourceIdentification &id, F &&func) const
{
    namespace proto = PRO_PROTO_NAMESPACE;

    proto::MacrosDocument_Macro *macro = engine_->CreateProto<proto::MacrosDocument_Macro>(step_);
    engine_->EnqueueResolveJob(
        step_,
        [id, macro](ClientChannel *channel) -> bool
        {
            ProUUID uuid;
            return TryParseProtoUUID(id.uuid, &uuid) && channel->RequestMacro(uuid, id.name, macro);
        },
        [func](
            TriggerEngine *engine,
            const std::shared_ptr<TriggerStep> step,
            std::weak_ptr<rendering::IRenderFactory> renderFactory,
            ClientChannel *channel)
        {
            ActionResolveContext context(engine, step, renderFactory, channel);
            func(&context);
        });
}

template<typename F>
void pro::ActionPrerollContext::RequestForceDisableBackground(F &&func) const
{
    ProPresentationColorTransfer *data = TriggerEngine::ArenaNew<ProPresentationColorTransfer>(m_step);

    m_engine->EnqueuePrerollJob(
        m_step,
        [data](ClientChannel *channel) -> bool
        {
            *data = channel->RequestForceDisableBackground();
            return true;
        },
        [data, func](
            TriggerEngine * /*engine*/,
            const std::shared_ptr<TriggerStep> /*step*/,
            std::weak_ptr<rendering::IRenderFactory> /*renderFactory*/,
            ClientChannel * /*channel*/) { func(*data); });
}

template<typename F>
void pro::ActionResolveContext::RequestStageDocument(F &&func) const
{
    namespace proto = PRO_PROTO_NAMESPACE;

    proto::Stage_Document *stageDoc = engine_->CreateProto<proto::Stage_Document>(step_);
    if (stageDoc == nullptr)
        return;

    engine_->EnqueueResolveJob(
        step_,
        [stageDoc](ClientChannel *channel) -> bool { return channel->RequestStageDocument(stageDoc); },
        [func](
            TriggerEngine *engine,
            const std::shared_ptr<TriggerStep> step,
            std::weak_ptr<rendering::IRenderFactory> renderFactory,
            ClientChannel *channel)
        {
            ActionPrerollContext context(engine, step, renderFactory);
            ResourceFactory factory(engine, step, renderFactory);
            func(&context, &factory, channel);
        });
}

template<typename F>
void pro::ActionPrerollContext::RequestPropCue(const ProUUID &id, const std::string &name, F &&func) const
{
    namespace proto = PRO_PROTO_NAMESPACE;

    struct RequestData
    {
        ProTriggerHandle handle;
        proto::Cue *cue;
        void *clientData;
    };

    auto op = m_step->GetOwningOperation();

    RequestData *data = m_engine->ArenaNew<RequestData>(m_step);
    data->handle = op->GetHandle();
    data->cue = m_engine->CreateProto<proto::Cue>(m_step);

    m_engine->EnqueuePrerollJob(
        m_step,
        [id, name, data](ClientChannel *channel) -> bool
        { return channel->RequestPropCue(id, name, data->handle, data->cue, &data->clientData); },
        [func](
            TriggerEngine *engine,
            const std::shared_ptr<TriggerStep> step,
            std::weak_ptr<rendering::IRenderFactory> renderFactory,
            ClientChannel *channel)
        {
            ActionPrerollContext context(engine, step, renderFactory);
            ResourceFactory factory(engine, step, renderFactory);
            func(&context, &factory, channel);
        });
}

template<typename F>
void pro::ActionPrerollContext::RequestMessage(
    const PRO_PROTO_NAMESPACE::Action_MessageType *messageAction, F &&func) const
{
    namespace proto = PRO_PROTO_NAMESPACE;

    proto::Slide *slideProto = m_engine->CreateProto<proto::Slide>(m_step);
    proto::Message *messageProto = m_engine->CreateProto<proto::Message>(m_step);

    if (slideProto == nullptr)
        return;

    m_engine->EnqueuePrerollJob(
        m_step,
        [messageAction, slideProto, messageProto](ClientChannel *channel) -> bool
        { return channel->RequestMessage(messageAction, slideProto, messageProto); },
        [slideProto, func](
            TriggerEngine *engine,
            const std::shared_ptr<TriggerStep> step,
            std::weak_ptr<rendering::IRenderFactory> renderFactory,
            ClientChannel *channel)
        {
            ActionPrerollContext context(engine, step, renderFactory);
            ResourceFactory factory(engine, step, renderFactory);
            func(&context, &factory, slideProto, channel);
        });
}

template<typename F>
void pro::ActionPrerollContext::RequestAltSlide(
    SlideHandle sourceSlide, AudienceLookHandle lookHandle, const ScreenLook &screenLook, F &&func) const
{
    namespace proto = PRO_PROTO_NAMESPACE;

    auto slideResource(m_engine->GetResource<triggering::resources::SlideResource>(sourceSlide));
    if (slideResource == nullptr)
        return;

    auto lookResource(m_engine->GetResource<triggering::resources::AudienceLookResource>(lookHandle));
    if (lookResource == nullptr)
        return;

    ProResourceAddress sourceAddress = slideResource->GetAddress();
    proto::Slide *altSlide = m_engine->CreateProto<proto::Slide>(m_step);
    bool *result = m_engine->ArenaNew<bool>(m_step);
    ProUUID lookID = lookResource->GetRequestID();

    m_engine->EnqueuePrerollJob(
        m_step,
        [lookID, screenLook, sourceAddress, altSlide, result](ClientChannel *channel) -> bool
        {
            *result = channel->RequestAltSlide(lookID, screenLook, sourceAddress, altSlide);
            return true;
        },
        [altSlide, func, result, lookID, screenLook](
            TriggerEngine *engine,
            const std::shared_ptr<TriggerStep> step,
            std::weak_ptr<rendering::IRenderFactory> renderFactory,
            ClientChannel *channel)
        {
            if (*result)
            {
                ActionPrerollContext context(engine, step, renderFactory);
                ResourceFactory factory(engine, step, renderFactory);
                func(&context, &factory, altSlide, channel);
            }
            else
            {
                std::string error = "Error request for alt slide with look ID: " + UUIDToString(lookID) +
                                    " and screen ID: " + UUIDToString(screenLook.screenID);
                MessageLog::Error("%s", error.c_str());
            }
        });
}

// F signature: void func(ActionPrerollContext*, ResourceFactory*, const proto::Slide*)
template<typename F>
void pro::ActionPrerollContext::RequestMaskSlide(const ProUUID &maskID, const ProUUID &screenID, F &&func) const
{
    namespace proto = PRO_PROTO_NAMESPACE;

    proto::Slide *maskSlidePtr = m_engine->CreateProto<proto::Slide>(m_step);
    bool *result = m_engine->ArenaNew<bool>(m_step);

    m_engine->EnqueuePrerollJob(
        m_step,
        [maskID, maskSlidePtr, result](ClientChannel *ch) -> bool
        {
            *result = ch->RequestMaskSlide(maskID, maskSlidePtr);

            return true;
        },
        [func](
            TriggerEngine *engine,
            const std::shared_ptr<TriggerStep> step,
            std::weak_ptr<rendering::IRenderFactory> rf,
            ClientChannel *channel)
        {
            ActionPrerollContext context(engine, step, rf);
            ResourceFactory factory(engine, step, rf);
            func(&context, &factory, channel);
        });
}

template<typename F>
void pro::ActionPrerollContext::RequestCCLISlide(SlideHandle sourceSlideHandle, F &&func) const
{
    namespace proto = PRO_PROTO_NAMESPACE;

    auto slideResource(m_engine->GetResource<triggering::resources::SlideResource>(sourceSlideHandle));

    if (slideResource == nullptr)
        return;

    ProResourceAddress sourceAddress = slideResource->GetAddress();
    proto::Slide *ccliSlide = m_engine->CreateProto<proto::Slide>(m_step);
    bool *result = m_engine->ArenaNew<bool>(m_step);

    // PRO-15393 - This may be a bit confusing and warrants a bit of explanation...  We are intentionally sending the
    // height/width for the slide, not the screen. This is mainly being driven by the Mac client. On the Mac client, the
    // Alt slide(alternateSlideHandler) is rendered where the size(slide size vs screen size) is determined based on
    // whether or not the slide has a look-template. When it is time for the _ccliSlide to run, it gets its size from
    // the transfer request (which was normally screen size). This causes a mismatch in element coordinates and thus the
    // moving CCLI text. Currently both clients composite the elements together using the slide coordinates and then
    // up/down scale as needed to fit the screen.
    uint32_t screenWidth, screenHeight;
    ProSize size = slideResource->Slide()->Size();
    screenWidth = (uint32_t)size.width;
    screenHeight = (uint32_t)size.height;

    m_engine->EnqueuePrerollJob(
        m_step,
        [screenWidth, screenHeight, sourceAddress, ccliSlide, result](ClientChannel *channel) -> bool
        {
            *result = channel->RequestCCLISlide(screenWidth, screenHeight, sourceAddress, ccliSlide);

            return true;
        },
        [ccliSlide, func](
            TriggerEngine *engine,
            const std::shared_ptr<TriggerStep> step,
            std::weak_ptr<rendering::IRenderFactory> renderFactory,
            ClientChannel *channel)
        {
            ActionPrerollContext context(engine, step, renderFactory);
            ResourceFactory factory(engine, step, renderFactory);
            func(&context, &factory, ccliSlide, channel);
        });
}

template<typename F>
void pro::ActionPrerollContext::RequestTimer(
    const ProUUID &timerID,
    const std::string &timerName,
    const PRO_PROTO_NAMESPACE::Timer_Configuration *timerConfiguration,
    ProTimerActionType timerAction,
    float timerIncrementAmount,
    F &&func) const
{
    namespace proto = PRO_PROTO_NAMESPACE;

    proto::Timer *timerProto = m_engine->CreateProto<proto::Timer>(m_step);

    m_engine->EnqueuePrerollJob(
        m_step,
        [timerID, timerName, timerConfiguration, timerAction, timerIncrementAmount, timerProto](
            [[maybe_unused]] ClientChannel *channel) -> bool
        {
            if (auto timer = TimerHelpers::RequestTimer(
                    timerID, timerName, timerConfiguration, timerAction, timerIncrementAmount);
                timer.has_value())
            {
                *timerProto = timer.value();
                return true;
            }
            return false;
        },
        [timerProto, func](
            TriggerEngine *engine,
            const std::shared_ptr<TriggerStep> step,
            std::weak_ptr<rendering::IRenderFactory> renderFactory,
            ClientChannel *channel)
        {
            ActionPrerollContext context(engine, step, renderFactory);
            ResourceFactory factory(engine, step, renderFactory);
            func(&context, &factory, timerProto);
        });
}

template<typename F>
void pro::ActionPrerollContext::RequestPlaylistCue(
    ProPlaylistSourceType source, const ProUUID &id, const std::string &name, bool alwaysRetrigger, F &&func) const
{
    namespace proto = PRO_PROTO_NAMESPACE;

    struct RequestData
    {
        ProTriggerHandle handle;
        proto::Cue *cue;
        void *clientData;
    };

    RequestData *data = m_engine->ArenaNew<RequestData>(m_step);
    data->handle = m_engine->AcquireTriggerHandle();
    data->cue = m_engine->CreateProto<proto::Cue>(m_step);

    m_engine->EnqueuePrerollJob(
        m_step,
        [source, id, name, alwaysRetrigger, data](ClientChannel *channel) -> bool {
            return channel->RequestPlaylistCue(
                source, id, name, alwaysRetrigger, data->handle, data->cue, &data->clientData);
        },
        [data, func](
            TriggerEngine *engine,
            const std::shared_ptr<TriggerStep> step,
            std::weak_ptr<rendering::IRenderFactory> renderFactory,
            ClientChannel *channel)
        {
            ActionPrerollContext context(engine, step, renderFactory);
            ResourceFactory factory(engine, step, renderFactory);
            func(&context, &factory, data->handle, data->cue, data->clientData);
        });
}

template<typename T>
std::shared_ptr<T> pro::ActionPrerollContext::GetResource(TypedActionResourceHandle<T::static_type> handle)
{
    return m_engine->GetResource<T>(handle);
}

template<typename T>
std::shared_ptr<T> pro::ActionPrerollContext::GetActiveResource(ProLayerIndex layer, ProUUID screenID) const
{
    return m_engine->GetActiveResource<T>(layer, screenID);
}
